@@ -41,7 +41,9 @@ enum {
 	THEME_GNOME,
 	THEME_GTK,
 	THEME_ENGINE,
-	THEME_METACITY
+	THEME_METACITY,
+	THEME_CURSOR,
+	THEME_ICON_CURSOR
 };
 
 enum {
@@ -67,15 +69,13 @@ static int
 file_theme_type (const gchar *dir)
 {
 	gchar *filename = NULL;
-	GnomeVFSURI *src_uri = NULL;
 	gboolean exists;
 
-	if (!dir) return THEME_INVALID;
+	if (!dir)
+		return THEME_INVALID;
 
 	filename = g_build_filename (dir, "index.theme", NULL);
-	src_uri = gnome_vfs_uri_new (filename);
-	exists = gnome_vfs_uri_exists (src_uri);
-	gnome_vfs_uri_unref (src_uri);
+	exists = g_file_test (filename, G_FILE_TEST_IS_REGULAR);
 
 	if (exists) {
 		GPatternSpec *pattern = NULL;
@@ -85,6 +85,7 @@ file_theme_type (const gchar *dir)
 		gboolean match;
 
 		uri = gnome_vfs_get_uri_from_local_path (filename);
+		g_free (filename);
 		gnome_vfs_read_entire_file (uri, &file_size, &file_contents);
 		g_free (uri);
 
@@ -92,8 +93,24 @@ file_theme_type (const gchar *dir)
 		match = g_pattern_match_string (pattern, file_contents);
 		g_pattern_spec_free (pattern);
 
-		if (match)
-			return THEME_ICON;
+		if (match) {
+			pattern = g_pattern_spec_new ("*Directories=*");
+			match = g_pattern_match_string (pattern, file_contents);
+			g_pattern_spec_free (pattern);
+
+			if (match) {
+				/* check if we have a cursor, too */
+				filename = g_build_filename (dir, "cursors", NULL);
+				exists = g_file_test (filename, G_FILE_TEST_IS_DIR);
+				g_free (filename);
+
+				if (exists)
+					return THEME_ICON_CURSOR;
+				else
+					return THEME_ICON;
+			}
+			return THEME_CURSOR;
+		}
 
 		pattern = g_pattern_spec_new ("*[X-GNOME-Metatheme]*");
 		match = g_pattern_match_string (pattern, file_contents);
@@ -101,32 +118,35 @@ file_theme_type (const gchar *dir)
 
 		if (match)
 			return THEME_GNOME;
+	} else {
+		g_free (filename);
 	}
-	g_free (filename);
 
 	filename = g_build_filename (dir, "gtk-2.0", "gtkrc", NULL);
-	src_uri = gnome_vfs_uri_new (filename);
+	exists = g_file_test (filename, G_FILE_TEST_IS_REGULAR);
 	g_free (filename);
-	exists = gnome_vfs_uri_exists (src_uri);
-	gnome_vfs_uri_unref (src_uri);
 
 	if (exists)
 		return THEME_GTK;
 
 	filename = g_build_filename (dir, "metacity-1", "metacity-theme-1.xml", NULL);
-	src_uri = gnome_vfs_uri_new (filename);
+	exists = g_file_test (filename, G_FILE_TEST_IS_REGULAR);
 	g_free (filename);
-	exists = gnome_vfs_uri_exists (src_uri);
-	gnome_vfs_uri_unref (src_uri);
 
 	if (exists)
 		return THEME_METACITY;
 
+	/* cursor themes don't necessarily have an index.theme */
+	filename = g_build_filename (dir, "cursors", NULL);
+	exists = g_file_test (filename, G_FILE_TEST_IS_DIR);
+	g_free (filename);
+
+	if (exists)
+		return THEME_CURSOR;
+
 	filename = g_build_filename (dir, "configure", NULL);
-	src_uri = gnome_vfs_uri_new (filename);
+	exists = g_file_test (filename, G_FILE_TEST_IS_EXECUTABLE);
 	g_free (filename);
-	exists = gnome_vfs_uri_exists (src_uri);
-	gnome_vfs_uri_unref (src_uri);
 
 	if (exists)
 		return THEME_ENGINE;
@@ -214,7 +234,7 @@ transfer_done_archive (gint filetype, const gchar *tmp_dir, const gchar *archive
 }
 
 static gboolean
-gnome_theme_install_real (gint filetype, const gchar *tmp_dir, const gchar *theme_name)
+gnome_theme_install_real (gint filetype, const gchar *tmp_dir, const gchar *theme_name, gboolean ask_user)
 {
 	gboolean success = TRUE;
 	GtkWidget *dialog, *apply_button;
@@ -224,38 +244,44 @@ gnome_theme_install_real (gint filetype, const gchar *tmp_dir, const gchar *them
 	gchar *user_message = NULL;
 	gchar *target_dir = NULL;
 
-	/* What type of theme it is ? */
+	/* What type of theme is it? */
 	theme_type = file_theme_type (tmp_dir);
-	if (theme_type == THEME_ICON) {
+	switch (theme_type) {
+	case THEME_ICON:
+	case THEME_CURSOR:
+	case THEME_ICON_CURSOR:
 		target_dir = g_build_path (G_DIR_SEPARATOR_S,
 					   g_get_home_dir (), ".icons",
 					   theme_name, NULL);
-	} else if (theme_type == THEME_GNOME) {
+		break;
+	case THEME_GNOME:
 		target_dir = g_build_path (G_DIR_SEPARATOR_S,
 					   g_get_home_dir (), ".themes",
 			 		   theme_name, NULL);
 		user_message = g_strdup_printf (_("GNOME Theme %s correctly installed"),
 						theme_name);
-	} else if (theme_type == THEME_METACITY ||
-		   theme_type == THEME_GTK) {
+		break;
+	case THEME_METACITY:
+	case THEME_GTK:
 		target_dir = g_build_path (G_DIR_SEPARATOR_S,
 					   g_get_home_dir (), ".themes",
 					   theme_name, NULL);
-	} else if (theme_type == THEME_ENGINE) {
+		break;
+	case THEME_ENGINE:
 		dialog = gtk_message_dialog_new (NULL,
 			       GTK_DIALOG_MODAL,
 			       GTK_MESSAGE_ERROR,
 			       GTK_BUTTONS_OK,
-			       _("The theme is an engine. You need to compile it."));
+			       _("The selected file is not a valid theme. It might be a theme engine that you need to compile."));
 		gtk_dialog_run (GTK_DIALOG (dialog));
 		gtk_widget_destroy (dialog);
 		return FALSE;
-	} else {
+	default:
 		dialog = gtk_message_dialog_new (NULL,
 			       GTK_DIALOG_MODAL,
 			       GTK_MESSAGE_ERROR,
 			       GTK_BUTTONS_OK,
-			       _("The file format is invalid"));
+			       _("The selected file does not appear to be a valid theme."));
 		gtk_dialog_run (GTK_DIALOG (dialog));
 		gtk_widget_destroy (dialog);
 		return FALSE;
@@ -277,15 +303,15 @@ gnome_theme_install_real (gint filetype, const gchar *tmp_dir, const gchar *them
 						 g_get_home_dir (),
 						 ".icons",
 						 theme_name, NULL);
-			update_icon_cache = g_strdup_printf ("gtk-update-icon-cache %s", new_path);
 			/* XXX: make some noise if we couldn't install it? */
 			gnome_vfs_move (path, new_path, FALSE);
 
 			/* update icon cache - shouldn't really matter if this fails */
+			update_icon_cache = g_strdup_printf ("gtk-update-icon-cache %s", new_path);
 			g_spawn_command_line_async (update_icon_cache, NULL);
+			g_free (update_icon_cache);
 
 			g_free (new_path);
-			g_free (update_icon_cache);
 		}
 		g_free (path);
 	}
@@ -302,16 +328,20 @@ gnome_theme_install_real (gint filetype, const gchar *tmp_dir, const gchar *them
 				GNOME_VFS_XFER_ERROR_MODE_ABORT,
 				GNOME_VFS_XFER_OVERWRITE_MODE_REPLACE,
 				NULL, NULL) != GNOME_VFS_OK) {
+		gchar *str;
+
+		str = g_strdup_printf (_("Installation for theme \"%s\" failed."), theme_name);
 		dialog = gtk_message_dialog_new (NULL,
 					GTK_DIALOG_MODAL,
 					GTK_MESSAGE_ERROR,
 					GTK_BUTTONS_OK,
-					_("Installation Failed"));
+					str);
+		g_free (str);
 		gtk_dialog_run (GTK_DIALOG (dialog));
 		gtk_widget_destroy (dialog);
 		success = FALSE;
 	} else {
-		if (theme_type == THEME_ICON)
+		if (theme_type == THEME_ICON || theme_type == THEME_ICON_CURSOR)
 		{
 			gchar *update_icon_cache;
 
@@ -321,65 +351,76 @@ gnome_theme_install_real (gint filetype, const gchar *tmp_dir, const gchar *them
 
 			g_free (update_icon_cache);
 		}
-		/* Ask to apply theme (if we can) */
-		if (theme_type == THEME_GTK || theme_type == THEME_METACITY || theme_type == THEME_ICON)
+
+		if (ask_user)
 		{
-			/* TODO: currently cannot apply "gnome themes" */
-			gchar *str;
-
-			str = g_strdup_printf(_("The theme \"%s\" has been installed."), theme_name);
-			user_message = g_strdup_printf("<span weight=\"bold\" size=\"larger\">%s</span>", str);
-			g_free (str);
-
-			dialog = gtk_message_dialog_new_with_markup (NULL, GTK_DIALOG_MODAL, GTK_MESSAGE_INFO, GTK_BUTTONS_NONE, user_message);
-
-			gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (dialog), _("Would you like to apply it now, or keep your current theme?"));
-
-			gtk_dialog_add_button (GTK_DIALOG (dialog), _("Keep Current Theme"), GTK_RESPONSE_CLOSE);
-
-			apply_button = gtk_button_new_with_label (_("Apply New Theme"));
-			gtk_button_set_image (GTK_BUTTON (apply_button), gtk_image_new_from_stock (GTK_STOCK_APPLY, GTK_ICON_SIZE_BUTTON));
-			gtk_dialog_add_action_widget (GTK_DIALOG (dialog), apply_button, GTK_RESPONSE_APPLY);
-			GTK_WIDGET_SET_FLAGS (apply_button, GTK_CAN_DEFAULT);
-			gtk_widget_show (apply_button);
-
-			gtk_dialog_set_default_response (GTK_DIALOG (dialog), GTK_RESPONSE_APPLY);
-
-			if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_APPLY)
+			/* Ask to apply theme (if we can) */
+			if ((theme_type == THEME_GTK || theme_type == THEME_METACITY ||
+			    theme_type == THEME_ICON || theme_type == THEME_CURSOR ||
+			    theme_type == THEME_ICON_CURSOR) && ask_user)
 			{
-				/* apply theme here! */
-				GConfClient *gconf_client;
-				const gchar *gconf_key;
+				/* TODO: currently cannot apply "gnome themes" */
+				gchar *str;
 
-				switch (theme_type)
+				str = g_strdup_printf (_("The theme \"%s\" has been installed."), theme_name);
+				user_message = g_strdup_printf ("<span weight=\"bold\" size=\"larger\">%s</span>", str);
+				g_free (str);
+
+				dialog = gtk_message_dialog_new_with_markup (NULL, GTK_DIALOG_MODAL, GTK_MESSAGE_INFO, GTK_BUTTONS_NONE, user_message);
+
+				gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (dialog), _("Would you like to apply it now, or keep your current theme?"));
+
+				gtk_dialog_add_button (GTK_DIALOG (dialog), _("Keep Current Theme"), GTK_RESPONSE_CLOSE);
+
+				apply_button = gtk_button_new_with_label (_("Apply New Theme"));
+				gtk_button_set_image (GTK_BUTTON (apply_button), gtk_image_new_from_stock (GTK_STOCK_APPLY, GTK_ICON_SIZE_BUTTON));
+				gtk_dialog_add_action_widget (GTK_DIALOG (dialog), apply_button, GTK_RESPONSE_APPLY);
+				GTK_WIDGET_SET_FLAGS (apply_button, GTK_CAN_DEFAULT);
+				gtk_widget_show (apply_button);
+
+				gtk_dialog_set_default_response (GTK_DIALOG (dialog), GTK_RESPONSE_APPLY);
+
+				if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_APPLY)
 				{
-					case THEME_GTK:
-						gconf_key = GTK_THEME_KEY;
-						break;
-					case THEME_METACITY:
-						gconf_key = METACITY_THEME_KEY;
-						break;
-					case THEME_ICON:
-						gconf_key = ICON_THEME_KEY;
-						break;
-					default: /* keep gcc happy */
-						gconf_key = NULL;
-						break;
-				}
+					/* apply theme here! */
+					GConfClient *gconf_client;
 
-				gconf_client = gconf_client_get_default ();
-				gconf_client_set_string (gconf_client, gconf_key, theme_name, NULL);
-				g_object_unref (gconf_client);
+					gconf_client = gconf_client_get_default ();
+
+					switch (theme_type)
+					{
+						case THEME_GTK:
+							gconf_client_set_string (gconf_client, GTK_THEME_KEY, theme_name, NULL);
+							break;
+						case THEME_METACITY:
+							gconf_client_set_string (gconf_client, METACITY_THEME_KEY, theme_name, NULL);
+							break;
+						case THEME_ICON:
+							gconf_client_set_string (gconf_client, ICON_THEME_KEY, theme_name, NULL);
+							break;
+						case THEME_CURSOR:
+							gconf_client_set_string (gconf_client, CURSOR_THEME_KEY, theme_name, NULL);
+							break;
+						case THEME_ICON_CURSOR:
+							gconf_client_set_string (gconf_client, ICON_THEME_KEY, theme_name, NULL);
+							gconf_client_set_string (gconf_client, CURSOR_THEME_KEY, theme_name, NULL);
+							break;
+						default:
+							break;
+					}
+
+					g_object_unref (gconf_client);
+				}
+			} else {
+				dialog = gtk_message_dialog_new (NULL,
+				       GTK_DIALOG_MODAL,
+				       GTK_MESSAGE_INFO,
+				       GTK_BUTTONS_OK,
+				       user_message);
+				gtk_dialog_run (GTK_DIALOG (dialog));
 			}
-		} else {
-			dialog = gtk_message_dialog_new (NULL,
-			       GTK_DIALOG_MODAL,
-			       GTK_MESSAGE_INFO,
-			       GTK_BUTTONS_OK,
-			       user_message);
-			gtk_dialog_run (GTK_DIALOG (dialog));
+			gtk_widget_destroy (dialog);
 		}
-		gtk_widget_destroy (dialog);
 	}
 
 	g_free (user_message);
@@ -410,7 +451,7 @@ transfer_done_cb (GtkWidget *dlg, gchar *path)
 						GTK_DIALOG_MODAL,
 						GTK_MESSAGE_ERROR,
 						GTK_BUTTONS_OK,
-						_("This theme is not in a supported format."));
+						_("The selected file does not appear to be a valid theme."));
 		gtk_dialog_run (GTK_DIALOG (dialog));
 		gtk_widget_destroy (dialog);
 		g_free (path);
@@ -419,7 +460,7 @@ transfer_done_cb (GtkWidget *dlg, gchar *path)
 
 	if (filetype == DIRECTORY) {
 		gchar *name = g_path_get_basename (path);
-		gnome_theme_install_real (filetype, path, name);
+		gnome_theme_install_real (filetype, path, name, TRUE);
 		g_free (name);
 	} else {
 		/* Create a temp directory and uncompress file there */
@@ -427,6 +468,7 @@ transfer_done_cb (GtkWidget *dlg, gchar *path)
 		const gchar *name;
 		gchar *tmp_dir;
 		gboolean ok;
+		gint n_themes;
 
 		tmp_dir = g_strdup_printf ("%s/.themes/.theme-%u",
 					   g_get_home_dir (),
@@ -455,22 +497,51 @@ transfer_done_cb (GtkWidget *dlg, gchar *path)
 
 		gnome_vfs_unlink (path);
 
+		/* See whether we have multiple themes to install. If so,
+		 * we won't ask the user whether to apply the new theme
+		 * after installation. */
+		n_themes = 0;
+		for (name = g_dir_read_name (dir); name && n_themes <= 1;
+		     name = g_dir_read_name (dir))
+		{
+			gchar *theme_dir;
+
+			theme_dir = g_build_filename (tmp_dir, name, NULL);
+
+			if (g_file_test (theme_dir, G_FILE_TEST_IS_DIR))
+				++n_themes;
+
+			g_free (theme_dir);
+		}
+		g_dir_rewind (dir);
+
 		ok = TRUE;
 		for (name = g_dir_read_name (dir); name && ok;
 		     name = g_dir_read_name (dir))
 		{
 			gchar *theme_dir;
 
-			theme_dir = g_build_path (G_DIR_SEPARATOR_S,
-						  tmp_dir, name, NULL);
+			theme_dir = g_build_filename (tmp_dir, name, NULL);
 
 			if (g_file_test (theme_dir, G_FILE_TEST_IS_DIR))
-				ok = gnome_theme_install_real (filetype, theme_dir, name);
+				ok = gnome_theme_install_real (filetype,
+					theme_dir, name, n_themes == 1);
 
 			g_free (theme_dir);
 		}
 		g_dir_close (dir);
 
+		if (ok && n_themes > 1)
+		{
+			dialog = gtk_message_dialog_new (NULL,
+			       GTK_DIALOG_MODAL,
+			       GTK_MESSAGE_INFO,
+			       GTK_BUTTONS_OK,
+			       _("New themes have been successfully installed."));
+			gtk_dialog_run (GTK_DIALOG (dialog));
+			gtk_widget_destroy (dialog);
+		}
+
 		cleanup_tmp_dir (tmp_dir);
 		g_free (tmp_dir);
 	}
@@ -555,7 +626,7 @@ gnome_theme_install_from_uri (const gchar *filename, GtkWindow *parent)
 					GTK_DIALOG_MODAL,
 					GTK_MESSAGE_ERROR,
 					GTK_BUTTONS_OK,
-					_("The file format is invalid."));
+					_("The selected file does not appear to be a valid theme."));
 		gtk_dialog_run (GTK_DIALOG (dialog));
 		gtk_widget_destroy (dialog);
 		g_free (base);
@@ -611,6 +682,7 @@ gnome_theme_installer_run (GtkWindow *parent, const gchar *filename)
 		filename = old_folder;
 
 	dialog = gtk_file_chooser_dialog_new (_("Select Theme"), parent, GTK_FILE_CHOOSER_ACTION_OPEN, GTK_STOCK_CANCEL, GTK_RESPONSE_REJECT, GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT, NULL);
+	gtk_dialog_set_default_response (GTK_DIALOG (dialog), GTK_RESPONSE_ACCEPT);
 
 	filter = gtk_file_filter_new ();
 	gtk_file_filter_set_name (filter, _("Theme Packages"));
