@@ -209,7 +209,7 @@ static TranslationEntry translations [] = {
 	  GCONF_VALUE_STRING,	translate_string_string },
 	{ "/desktop/gnome/interface/toolbar_style",			"Gtk/ToolbarStyle",
 	  GCONF_VALUE_STRING,	translate_string_string_toolbar },
-	{ "/desktop/gnome/interface/toolbar_icon_size",		"Gtk/ToolbarIconSize",
+	{ "/desktop/gnome/interface/toolbar_icons_size",		"Gtk/ToolbarIconSize",
 	  GCONF_VALUE_STRING,	translate_string_string },
 	{ "/desktop/gnome/interface/can_change_accels",		"Gtk/CanChangeAccels",
 	  GCONF_VALUE_BOOL,		translate_bool_int },
